/*
** Module : StreamFactory.java
** Public : creates stream instances on a remote node
**
** Copyright (c) 2006-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20060211   @24554 Created initial version which creates file and process stream
**                           instances on a remote node.
** 002 SVL 20080219   @37121 openFileStream() returns null if filename equals "terminal".
** 003 SIY 20080317   @37456 Added support for terminal stream. Ref: #37453
** 004 SIY 20090331   @41707 Added more openFileStream() signatures and added detection of 
**                           interactive terminal redirection. Ref: #41705
** 005 SIY 20090401   @41714 Fixed regression introduced by #004: wrong order of page size
**                           setting calls. 
** 006 SIY 20090401   @41715 Do not call setPageSiz() if NO_PAGING is passed as a page size and 
**                           paging flag is set to false. Yet another regression from #004. 
** 007 GES 20111020          Modified per-session initialization to detect when the client code
**                           is running inside the server process AND to initialize for local
**                           redirection instead of using a remote object.
** 008 CS  20130103          Implement OS-DIR support as an INPUT stream
** 009 VIG 20120918          Added printer stream support for OUTPUT TO PRINTER statement.
** 010 EVK 20140601          Empty directory name should be current directory ".".
** 011 GES 20140612          Added explicit support for the terminal stream. 
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.ContextLocal;

/**
 * Creates file and process stream instances that are on the remote node 
 * associated with the current user's session and which are accessed via a
 * {@link RemoteStream} instance.
 */
public class StreamFactory
{
   /** Special page size which causes turning paging off.*/
   private static final int NO_PAGING = -2;
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /**
    * Constructs a remote stream instance representing a file using a filename, this file or
    * device will be opened for reading or writing based on the given <code>write</code> flag.
    * The current read/write position will be at offset 0 unless in <code>write</code> mode and 
    * the <code>append</code> flag is on (in which case the write position will be at the end of
    * the file).
    *
    * @param    filename
    *           The name of the file or device to open.  Must not be <code>null</code>.
    * @param    write
    *           If <code>true</code>, open in write mode, otherwise open the file for reading.
    * @param    append
    *           In <code>write</code> mode, if <code>true</code>, the current file pointer is
    *           set to the end of the file. Otherwise if in <code>write</code> mode, the file
    *           is truncated to 0 length. This value is ignored in <code>read</code> mode.
    * @param    pageSz
    *           Desired page size. <code>-1</code> means default paging.
    * @param    paged
    *           <code>true</code> means that stream is paged. 
    *           
    * @return   The new stream instance or <code>null</code> if an error occurred while silent
    *           error mode was active.
    *
    * @throws   ErrorConditionException
    *           If the file does not exist OR if the current user does not have the proper rights
    *           to open the file or device.
    */
   public static Stream openFileStream(String  filename,
                                       boolean write,
                                       boolean append,
                                       int     pageSz,
                                       boolean paged)
   throws ErrorConditionException
   {
      int id = work.obtain().rsb.openFileStream(filename, write, append);
      
      Stream res = (id < 0) ? null : new RemoteStream(id);
      
      if (res != null)
      {
         if ((pageSz < 0 && pageSz != NO_PAGING) || paged)
            res.setPageSize();

         if (pageSz >= 0)
            res.setPageSize(pageSz);
      }
      
      return res;
   }
   
   /**
    * Constructs a remote stream instance representing a file using a filename, this file or
    * device will be opened for reading or writing based on the given <code>write</code> flag.
    * The current read/write position will be at offset 0 unless in <code>write</code> mode and 
    * the <code>append</code> flag is on (in which case the write position will be at the end of
    * the file).
    *
    * @param    filename
    *           The name of the file or device to open.  Must not be <code>null</code>.
    * @param    write
    *           If <code>true</code>, open in write mode, otherwise open the file for reading.
    * @param    append
    *           In <code>write</code> mode, if <code>true</code>, the current file pointer is
    *           set to the end of the file. Otherwise if in <code>write</code> mode, the file
    *           is truncated to 0 length. This value is ignored in <code>read</code> mode.
    * @param    sz
    *           Desired page size. <code>-1</code> means default paging.
    *           
    * @return   The new stream instance or <code>null</code> if an error occurred while silent
    *           error mode was active.
    *
    * @throws   ErrorConditionException
    *           If the file does not exist OR if the current user does not have the proper rights
    *           to open the file or device.
    */
   public static Stream openFileStream(String filename, boolean write, boolean append, int sz)
   {
      return openFileStream(filename, write, append, sz, false);
   }
   
   /**
    * Constructs a remote stream instance representing a file using a filename, this file or
    * device will be opened for reading or writing based on the given <code>write</code> flag.
    * The current read/write position will be at offset 0 unless in <code>write</code> mode and 
    * the <code>append</code> flag is on (in which case the write position will be at the end of
    * the file).
    * <p>
    * This stream will NOT have paging active.
    *
    * @param    filename
    *           The name of the file or device to open.  Must not be <code>null</code>.
    * @param    write
    *           If <code>true</code>, open in write mode, otherwise open the file for reading.
    * @param    append
    *           In <code>write</code> mode, if <code>true</code>, the current file pointer is
    *           set to the end of the file. Otherwise if in <code>write</code> mode, the file
    *           is truncated to 0 length. This value is ignored in <code>read</code> mode.
    *           
    * @return   The new stream instance or <code>null</code> if an error occurred while silent
    *           error mode was active.
    *
    * @throws   ErrorConditionException
    *           If the file does not exist OR if the current user does not have the proper rights
    *           to open the file or device.
    */
   public static Stream openFileStream(String filename, boolean write, boolean append)
   {
      return openFileStream(filename, write, append, NO_PAGING, false);
   }

   /**
    * Constructs a remote stream instance representing a file using a filename, this file or
    * device will be opened for reading or writing based on the given <code>write</code> flag.
    * The current read/write position will be at offset 0 unless in <code>write</code> mode and 
    * the <code>append</code> flag is on (in which case the write position will be at the end of
    * the file).
    *
    * @param    filename
    *           The name of the file or device to open.  Must not be <code>null</code>.
    * @param    write
    *           If <code>true</code>, open in write mode, otherwise open the file for reading.
    * @param    append
    *           In <code>write</code> mode, if <code>true</code>, the current file pointer is
    *           set to the end of the file. Otherwise if in <code>write</code> mode, the file
    *           is truncated to 0 length. This value is ignored in <code>read</code> mode.
    * @param    paged
    *           <code>true</code> means that stream is paged with the default page size. 
    *           
    * @return   The new stream instance or <code>null</code> if an error occurred while silent
    *           error mode was active.
    *
    * @throws   ErrorConditionException
    *           If the file does not exist OR if the current user does not have the proper rights
    *           to open the file or device.
    */
   public static Stream openFileStream(String  filename,
                                       boolean write,
                                       boolean append,
                                       boolean paged)
   {
      return openFileStream(filename, write, append, NO_PAGING, paged);
   }
   
   /**
    * Constructs a remote stream instance representing a stream result of 
    * INPUT FROM OS-DIR statement.
    * 
    * @param  dir
    *         The <code>String</code> variable representing the directory for 
    *         which OS-DIR is applied.
    * @param  noAttrList
    *         This indicates if type index will no be output to stream If
    *         <code>true</code> the index will not be output to the stream,
    *         if <code>false</code> it will be output.
    *        
    * @return The <codeStream</code> wrapping the <code>OsDirStream</code> 
    *         containing the result of OS-DIR implementation, or null if
    *         an error has occurred.
    */
   public static Stream openDirStream(String dir, boolean noAttrList)
   {
      String fixDir = "".equals(dir) ? "." : dir;
      return new RemoteStream(work.obtain().rsb.openDirStream(fixDir,
                                                              noAttrList));
   }
   
   /**
    * Constructs a remote stream instance representing a stream result of 
    * INPUT FROM OS-DIR statement.
    * <p>
    * When no boolean variable representing the noAttrList which decides if 
    * file type will be added to the stream is present the file type will be 
    * added by default.
    * 
    * @param  dir
    *         The <code>String</code> variable representing the directory for 
    *         which OS-DIR is applied.
    *        
    * @return The <codeStream</code> wrapping the <code>OsDirStream</code> 
    *         containing the result of OS-DIR implementation, or null if
    *         an error has occurred.
    */
   public static Stream openDirStream(String dir)
   {
      return openDirStream(dir, false);
   }
   
   /**
    * Constructs a remote stream instance representing a stream result of 
    * INPUT FROM OS-DIR statement.
    * 
    * @param  dir
    *         The <code>character</code> variable representing the directory 
    *         for which OS-DIR is applied.
    * @param  noAttrList
    *         This indicates if type index will no be output to stream If
    *         <code>true</code> the index will not be output to the stream,
    *         if <code>false</code> it will be output.
    *        
    * @return The <codeStream</code> wrapping the <code>OsDirStream</code> 
    *         containing the result of OS-DIR implementation, or null if
    *         an error has occurred.
    */
   public static Stream openDirStream(character dir, boolean noAttrList)
   {
      return openDirStream(dir.toStringMessage(), noAttrList);
   }
   
   /**
    * Constructs a remote stream instance representing a stream result of 
    * INPUT FROM OS-DIR statement.
    * <p>
    * When no boolean variable representing the noAttrList which decides if 
    * file type will be added to the stream is present the file type will be 
    * added by default.
    * 
    * @param  dir
    *         The <code>character</code> variable representing the directory 
    *         for which OS-DIR is applied.
    *        
    * @return The <codeStream</code> wrapping the <code>OsDirStream</code> 
    *         containing the result of OS-DIR implementation, or null if
    *         an error has occurred.
    */
   public static Stream openDirStream(character dir)
   {
      return openDirStream(dir.toStringMessage(), false);
   }   

   /**
    * Constructs a remote stream instance representing a process' standard I/O.
    *
    * @return   The new stream instance.
    */
   public static Stream openProcessStream()
   {
      return new RemoteStream(work.obtain().rsb.openProcessStream());
   }
   
   /**
    * Constructs a remote stream instance representing the terminal.
    *
    * @return   The new stream instance.
    */
   public static Stream openTerminalStream()
   {
      return new RemoteStream(work.obtain().rsb.openTerminalStream(), true);
   }
   
   /**
    * Constructs a remote stream instance representing the printer Stream 
    * result of the OUTPUT TO PRINTER statement. 
    * It can be assinged as unnamed output (OUTPUT TO PRINTER statement)
    * or as named stream (OUTPUT STREAM someStream TO PRINTER)
    * 
    * @return   The stream id which is used to route operations to the
    *           newly created instance.
    */   
   public static Stream openPrinterStream()
   {
      return new RemoteStream(work.obtain().rsb.openPrinterStream());
   }
   
   /**
    * Constructs a remote stream instance representing the printer Stream 
    * result of the OUTPUT TO PRINTER statement. 
    * It can be assinged as unnamed output (OUTPUT TO PRINTER statement)
    * or as named stream (OUTPUT STREAM someStream TO PRINTER)
    * 
    * @param    printerName
    *           name of target printer
    * 
    * @return   The stream id which is used to route operations to the
    *           newly created instance.
    */   
   public static Stream openPrinterStream(String printerName)
   {
      return new RemoteStream(work.obtain().rsb.openPrinterStream(printerName));
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** The remote stream builder instance. */
      private StreamBuilder rsb = null;
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         // obtain the our proxy, server-side code is left-side by convention
         wa.rsb = (StreamBuilder)
            RemoteObject.obtainInstance(StreamBuilder.class, true);
                 
         return wa;
      }   
   }
}
